/***************************************************************************
                          floatrangefacttable.cpp  -  dtfarmer
                             -------------------
    begin                : Wed Sep 13 2000
    copyright            : (C) 2000 by Siegfried Nijssen
    email                : snijssen@cs.leidenuniv.nl
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include <values.h>
#include <stdio.h>
#include <math.h>
#include "output.h"
#include "floatrangefacttable.h"
#include "tree.h"
#include "globaldata.h"
#include "type.h"

struct MinMax {
  INTTYPE min, max;
};

enum TestType { LowerThan, HigherThan };
struct Test {
  double doublethreshold;
  double additionalpunishment; // MDL principle possibly allowed
  INTTYPE threshold;
  TestType type;
};

enum FRFTADType { DeterminingMinMax, FixedTest };
struct FRFTAdditionalData {
  FRFTADType type;
  union {
    MinMax *minmax;
    Test *test;
  } data;
};

FloatRangeFactTable::FloatRangeFactTable() {

}

FloatRangeFactTable::~FloatRangeFactTable() {

}

double FloatRangeFactTable::toDouble ( int i ) {
  Node *node = getCurrentNode ();
  ModeData *md = modetable.array + MODE ( node );
  int typenr = md->params[0].number;
  return typetable.getConstantDouble ( typenr, i );
}

double FloatRangeFactTable::additionalPunishment () {
  Node *node = getCurrentNode ();
  return ( (FRFTAdditionalData*) node->additionaldata )->data.test->additionalpunishment;
}

bool FloatRangeFactTable::getFact ( INTTYPE *coor ) {
  FRFTAdditionalData *frftad = (FRFTAdditionalData*) getCurrentNode ()->additionaldata;
  if ( frftad->type == DeterminingMinMax ) {
    if ( coor[0] < frftad->data.minmax->min )
      frftad->data.minmax->min = coor[0];
    if ( coor[0] > frftad->data.minmax->max )
      frftad->data.minmax->max = coor[0];
    return false; // push towards giving the next value
  }
  else
    if ( othertest )
      if ( frftad->data.test->type == LowerThan )
        return toDouble ( coor[0] ) <= frftad->data.test->doublethreshold;
      else
        return toDouble ( coor[0] ) > frftad->data.test->doublethreshold;
    else
      if ( frftad->data.test->type == LowerThan )
        return coor[0] <= frftad->data.test->threshold;
      else
        return coor[0] > frftad->data.test->threshold;
}

/************************************************************************/
/* ALGORITHM FOR DETERMINING BORDER                                     */
/************************************************************************/

class NumberClassArray {
  private:
    INTTYPE *array;
    int *classarray;
    int sizearray;
    KeyFactTables *kfts; // only used to determine sizes of various arrays
    int diff; // the number of different variables, might be used by the
              // MDL principle

    void split ( int left, int right, int &l1, int &r1, int &l2, int &r2 ) {
      int i = left, j = right + 1;
      int s = array[left], sclass = classarray[left];
      l1 = left;
      r2 = right;

      if ( i < j )
        while ( true ) {
          j--;
          while ( ( i != j ) && ( array[j] > s ) )
            j--;
          if ( i < j ) {
            array[i] = array[j];
            classarray[i] = classarray[j];
          }
          i++;
          while ( ( i < j ) && ( array[i] < s ) )
            i++;
          if ( i < j ) {
            array[j] = array[i];
            classarray[j] = classarray[i];
          }
          else
            break;
        }
      array[j] = s;
      classarray[j] = sclass;
      r1 = j - 1;
      l2 = j + 1;
    }

    void quicksort ( int left, int right ) {
      int l1, l2, r1, r2;

      if ( left < right ) {
        split ( left, right, l1, r1, l2, r2 );
        quicksort ( l1, r1 );
        quicksort ( l2, r2 );
      }
    }

    void quicksort () {
      quicksort ( 0, sizearray - 1 );
    }

    double impurity ( int i, int *classcnt ) {
      double t1 = 0.0, t2 = 0.0, t3, t4;
      int ta, tb, s;

      for ( int j = 0; j < kfts->size; j++ ) {
        ta = classcnt[j];
        s = kfts->keyfacttable[j] ? kfts->keyfacttable[j]->getSize () : 0;
        tb = s - ta;
        if ( ta )
          t1 += ( (double) ta ) * log ( ta );
        if ( tb )
          t2 += ( (double) tb ) * log ( tb );
      }

      t3 = i;
      t4 = sizearray - t3;
      return t3 * log ( t3 ) - t1 + t4 * log ( t4 ) - t2;
        // to obtain real impurity, divide by sizearray,
        // but as the final value is only used in comparisons,
        // and sizearray does not change, this suffices
    }

    int determine_split ( double &minimpurity, int &minsplit, int *classcnt ) {
      double t;
      int mini = 0;
      diff = 1;

      classcnt[classarray[0]] = 1;
      for ( int i = 1; i < sizearray; i++ ) {
        if ( array[i] != array[i - 1] ) {
          diff++;
          t = impurity ( i, classcnt );
          if ( t < minimpurity ) {
            minimpurity = t;
            mini = i - 1;
          }
        }
        classcnt[classarray[i]]++;
      }

      minsplit = array[mini];
      return array[mini + 1];
    }

  public:
    NumberClassArray ( KeyFactTables *kfts ):
      sizearray ( 0 ) {
      this->kfts = kfts;
      int maxsizearray = kfts->nrkeyfacts ();
      array = new INTTYPE[maxsizearray];
      classarray = new int[maxsizearray];
    }

    ~NumberClassArray () {
      delete[] array;
      delete[] classarray;
    }

    void add ( INTTYPE value, int clas ) {
      array[sizearray] = value;
      classarray[sizearray] = clas;
      sizearray++;
    }

    // minsplit: the integer index of the minimum impurity
    // returns the next integer index in the array,
    // used by the OTHER_TEST mechanism
    // mdl might be used for computation according to the mdl principle
    int findSplit ( double &minimpurity, int &minsplit, double &mdl ) {
      int ret;
      quicksort ();

      minimpurity = MAXDOUBLE;
      minsplit = 0;

      int classcnt[kfts->size];
      for ( int i = 0; i < kfts->size; i++ )
        classcnt[i] = 0;

      ret = determine_split ( minimpurity, minsplit, classcnt );

      mdl = log ( (double) diff ) / kfts->nrkeyfacts ();

      delete this;

      return ret;
    }
};

NumberClassArray *min, *max;

void fillin_minmax ( int cnr, void *additionaldata_class, INTTYPE *value, void *additionaldata_value, void *extraparams ) {
  MinMax *minmax = (MinMax*) additionaldata_value;

  min->add ( minmax->min, cnr );
  max->add ( minmax->max, cnr );
}

struct RedistributeParam {
  KeyFactTables *left, *right;
  Test *test;
};

void redistribute_value ( int cnr, void *additionaldata_class, INTTYPE *value, void *additionaldata_value, void *extraparams ) {
  bool left;
  RedistributeParam *param = (RedistributeParam*) extraparams;
  MinMax *minmax = (MinMax*) additionaldata_value;
  KeyFactTable **kft;

  if ( param->test->type == LowerThan )
    left = ( minmax->min <= param->test->threshold );
  else
    left = ( minmax->max > param->test->threshold );

  if ( left )
    kft = param->left->keyfacttable + cnr;
  else
    kft = param->right->keyfacttable + cnr;
  if ( !*kft )
    *kft = new KeyFactTable ( akeypredicate->size, false );
  (*kft)->addFact ( value, NULL );

  // not needed any more, is not removed by the destructor of the table
  // as it is a void pointer (originally, at least)
  delete minmax;
}

void FloatRangeFactTable::redistribute ( Test *test ) {
  RedistributeParam param;
  param.left = getCurrentNode ()->leftfacts;
  param.right = new KeyFactTables ( param.left->size );
  param.test = test;

  KeyFactTables *oldright = getCurrentNode ()->rightfacts;

  getCurrentNode ()->rightfacts = param.right;

  oldright->iterate ( redistribute_value, (void*) &param );

  delete oldright;
}

void FloatRangeFactTable::finish () {
  KeyFactTables *kfts = getCurrentNode ()->rightfacts;
  FRFTAdditionalData *frftad = (FRFTAdditionalData*) getCurrentNode ()->additionaldata;
  int minsplit, maxsplit, minnext, maxnext;
  double minimp, maximp, minmdl, maxmdl;

  min = new NumberClassArray ( kfts );
  max = new NumberClassArray ( kfts );

  kfts->iterate ( fillin_minmax, NULL );

  minnext = min->findSplit ( minimp, minsplit, minmdl );
  maxnext = max->findSplit ( maximp, maxsplit, maxmdl );

  /* transform the current node to a testing node */
  delete frftad->data.minmax;
  Test *test = new Test;
  frftad->type = FixedTest;
  frftad->data.test = test;

  /* determine test */
  bool bt;
  if ( mdl )
    bt = minimp + minmdl < maximp + maxmdl;
  else
    bt = minimp < maximp;

  if ( bt ) {
    if ( mdl )
      test->additionalpunishment = minmdl;
    else
      test->additionalpunishment = 0.0;
    test->type = LowerThan;
    test->threshold = minsplit;
    if ( othertest )
      test->doublethreshold = ( toDouble ( minsplit ) + toDouble ( minnext ) ) / 2;
  }
  else {
    if ( mdl )
      test->additionalpunishment = maxmdl;
    else
      test->additionalpunishment = 0.0;
    test->type = HigherThan;
    test->threshold = maxsplit;
    if ( othertest )
      test->doublethreshold = ( toDouble ( maxsplit ) + toDouble ( maxnext ) ) / 2;
  }

  /* redistribute the examples for this test */

  redistribute ( test );
}

/************************************************************************/

    /* this function is called by the constructor of a node, to
     * initialize the 'additionaldata' of the node. The default sets
     * additionaldata to NULL; derived classes can override
     */
void FloatRangeFactTable::init () {
  MinMax *minmax = new MinMax;
  minmax->min = MAXSHORT;
  minmax->max = 0;
  FRFTAdditionalData *frftad = new FRFTAdditionalData;
  frftad->type = DeterminingMinMax;
  frftad->data.minmax = minmax;
  getCurrentNode ()->additionaldata = (void*) frftad;
}

void FloatRangeFactTable::remove () {
  FRFTAdditionalData *frftad = (FRFTAdditionalData*) getCurrentNode ()->additionaldata;
  if ( frftad->type == DeterminingMinMax )
    delete frftad->data.minmax;
  else
    delete frftad->data.test;
  delete frftad;
}

    /* the keyfacttables have place for additional data for each example
     * separately. This function returns the information which is to be placed
     * in this location for this kind of table. Optionally uses a node to
     * determine the contents.
     * the default returns NULL.
     */
void *FloatRangeFactTable::determineAdditionalData () {
  FRFTAdditionalData *frftad = (FRFTAdditionalData*) getCurrentNode ()->additionaldata;
  MinMax *oldminmax = frftad->data.minmax,
         *newminmax = new MinMax;
  newminmax->min = MAXSHORT;
  newminmax->max = 0;
  frftad->data.minmax = newminmax;

  return (void*) oldminmax;
}

char *FloatRangeFactTable::string () {
  static char string[100];
  FRFTAdditionalData *frfta = (FRFTAdditionalData*) getCurrentNode ()->additionaldata;

  if ( frfta->type == DeterminingMinMax )
    sprintf ( string, "[%i,%i]", frfta->data.minmax->min, frfta->data.minmax->max );
  else {
    int tn = modetable.array[ MODE ( getCurrentNode () ) ].params[0].number;
    if ( !othertest )
      if ( frfta->data.test->type == LowerThan )
        sprintf ( string, "%s=>", typetable.getConstantName ( tn, frfta->data.test->threshold ) );
      else
        sprintf ( string, "%s<", typetable.getConstantName ( tn, frfta->data.test->threshold ) );
    else
      if ( frfta->data.test->type == LowerThan )
        sprintf ( string, "OT%f=>", frfta->data.test->doublethreshold );
      else
        sprintf ( string, "OT%f<", frfta->data.test->doublethreshold );
  }

  return string;
}
