/***************************************************************************
                          interpreter.cc  -  description
                             -------------------
    begin                : Tue Mar 21 2000
    copyright            : (C) 2000 by Siegfried Nijssen
    email                : snijssen@cs.leidenuniv.nl
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include <stdlib.h>
#include "type.h"
#include "interpreter.h"
#include "output.h"

bool parsingbias = true;

bool interpret_type ( int size, AtomString **params ) {
  char *type_name = params[0]->name;
  if ( size < 3 )
    return false;
  typetable.addType ( type_name, size - 1 );
  for ( int i = 1; i < size; i++ ) {
    if ( params[i]->nrofparams > 0 )
      return false;
    if ( !typetable.addConstant ( params[i]->name, type_name ) ) {
      ERR << "Illegal constant in type definition ";
      return false;
    }
  }
  return true;
}

bool interpret_inttype ( int size, AtomString **params ) {
  if ( size != 2 )
    return false;
  char *type_name = params[0]->name;
  int n = atoi ( params[1]->name );
  typetable.addType ( type_name, n );
  return true;
}

bool interpret_predicate ( int size, AtomString **params ) {
  if ( size != 1 )
    return false;
  char *name = params[0]->name;
  size = params[0]->nrofparams;
  int *paramtypes = new int[size], typenr, nrofconst;
  params = params[0]->params;
  for ( int i = 0; i < size; i++ ) {
    if ( params[i]->nrofparams != 0 )
      return false;
    typetable.getType ( params[i]->name, typenr, nrofconst );
    paramtypes[i] = typenr;
  }
  predicatetable.addPredicate ( name, size, paramtypes );
  
  return true;
}

bool interpret_mode ( int size, AtomString **params, char *functionname = NULL ) {
  if ( size != 2 && size != 3 )
    return false;
  int number = atoi ( params[0]->name ), typenr, constantnr, idnr;
  char *name = params[1]->name;
  if ( size == 3 )
    idnr = atoi ( params[2]->name );
  else
    idnr = -1;
  PredicateData *predicatedata = predicatetable.getPredicate ( name );
  if ( !predicatedata ) {
    ERR << "Unknown predicate " << name << ": ";
    return false;
  }
  
  int predicate = predicatedata->nr,
      predsize = predicatedata->size;
  if ( params[1]->nrofparams != predsize ) {
    ERR << "Wrong number of parameters: ";
    return false;
  }
  ModeParamData *modeparams = new ModeParamData[predsize];
  AtomString **walk = params[1]->params;

  for ( int i = 0; i < predsize; i++ )
    switch ( walk[i]->name[0] ) {
      case '+': {
        modeparams[i].type = Plus;
        modeparams[i].number = predicatedata->params[i];
        break;
      }
      case '-': {
        modeparams[i].type = Min;
	modeparams[i].number = predicatedata->params[i];
	break;
      }
      case '\\': {
        modeparams[i].type = MinEx;
        modeparams[i].number = predicatedata->params[i];
        break;
      }
      default: {
        modeparams[i].type = Constant;
        typenr = predicatedata->params[i];
          // define the type expected, could be used in the
          // search for the number constantnr
	typetable.getConstant ( walk[i]->name, typenr, constantnr );
	if ( constantnr == -1 ) {
	  ERR << "Unknown constant " << walk[i]->name << ": ";
	  return false;
	}
	if ( typenr != -1 && typenr != predicatedata->params[i] ) {
	  ERR << "Constant " << walk[i]->name << " is of wrong type: ";
	  return false;
	}
	modeparams[i].number = constantnr;
      }
    }

  modetable.addMode ( number, predicate, predsize, modeparams, idnr, functionname );
  
  return true;
}

bool interpret_fact ( char *name, int size, AtomString **params ) {
  PredicateData *predicatedata;
  if ( name[0] <= '9' )
    predicatedata = predicatetable.getPredicate ( atoi ( name ) );
  else
    predicatedata = predicatetable.getPredicate ( name );
  int typenr, constantnr;

  if ( !predicatedata ) {
    ERR << "Unknown predicate used in fact: ";
    return false;
  }
  if ( size != predicatedata->size ) {
    ERR << "Incorrect number of parameters: ";
    return false;
  }
  // do not test for key: keys should have facts!
  INTTYPE coor[size];
  for ( int i = 0; i < size; i++ ) {
    typenr = predicatedata->params[i];
    typetable.getConstant ( params[i]->name, typenr, constantnr );
    if ( typenr != -1 && ( typenr != predicatedata->params[i] || constantnr == -1 ) ) {
      ERR << "Constant " << params[i]->name << " is of incorrect type: ";
      return false;
    }
    coor[i] = constantnr;
  }
  setFact ( predicatedata, coor );
  return true;
}

bool interpret_key ( int size, AtomString **params ) {
  PredicateData *pd, *pdf = NULL;

  for ( int i = 0; i < size; i++ ) {
    akeypredicate = pd = predicatetable.getPredicate ( params[i]->name );
    if ( !pd ) {
      cerr << "Unknown predicate " << params[0]->name << ": ";
      return false;
    }
    if ( pdf == NULL )
      pdf = pd;
    else {
      for ( int j = 0; j < pd->size; j++ )
        if ( pd->params[j] != pdf->params[j] ) {
          cerr << "Not all predicates for the key have the same parameter types!" << endl;
          return false;
        }
    }
    
    pd->key = modetable.keynr;

    // add as mode declaration:
    int predsize = pd->size;
    ModeParamData *modeparams = new ModeParamData[predsize];

    for ( int i = 0; i < predsize; i++ ) {
      modeparams[i].type = Min;
      modeparams[i].number = pd->params[i];
    }

    modetable.addMode ( 1, pd->nr, predsize, modeparams, -1, NULL, true );
  }

  return true;
}

bool interpret_showconstantnames ( int size, AtomString **params ) {
  if ( size != 1 ) {
    ERR << "Show constant names needs one parameter: ";
    return false;
  }
  if ( strcmp ( params[0]->name, "true" ) == 0 )
    showconstantnames = true;
  else
    if ( strcmp ( params[0]->name, "false" ) == 0 )
      showconstantnames = false;
    else {
      ERR << "Show constant names parameter should be true or false ";
      return false;
    }
  return true;
}

bool interpret_numberimm ( int size, AtomString **params ) {
  if ( size != 1 ) {
    ERR << "Numberimm needs one parameter: ";
    return false;
  }
  if ( strcmp ( params[0]->name, "true" ) == 0 )
    typetable.set_number_imm ( true );
  else
    if ( strcmp ( params[0]->name, "false" ) == 0 )
      typetable.set_number_imm ( false );
    else {
      ERR << "Numberimm parameter should be true or false ";
      return false;
    }
  return true;
}

bool interpret_intoption ( int size, AtomString **params, char *name, int &option ) {
  if ( size != 1 ) {
    ERR << name << " needs one parameter: ";
    return false;
  }
  option = atoi ( params[0]->name );
  return true;
}

bool interpret_doubleoption ( int size, AtomString **params, char *name, double &option ) {
  if ( size != 1 ) {
    ERR << name << " needs one parameter: ";
    return false;
  }
  option = atof ( params[0]->name );
  return true;
}

bool interpret_booloption ( int size, AtomString **params, char *name, bool &option ) {
  if ( size != 1 ) {
    ERR << name << " needs one parameter: ";
    return false;
  }
  option = strcmp ( params[0]->name, "false" );
  return true;
}

#define INTERPRET_INTOPTION(atom,_name) \
  if ( strcmp ( atom->name, #_name ) == 0 ) \
    return interpret_intoption ( atom->nrofparams, atom->params, #_name, _name );
#define INTERPRET_DOUBLEOPTION(atom,_name) \
  if ( strcmp ( atom->name, #_name ) == 0 ) \
    return interpret_doubleoption ( atom->nrofparams, atom->params, #_name, _name );
#define INTERPRET_BOOLOPTION(atom,_name) \
  if ( strcmp ( atom->name, #_name ) == 0 )\
    return interpret_booloption ( atom->nrofparams, atom->params, #_name, _name );


void fill () {
  ERR << "[Processing bias]" << endl;
  ERR_F
  typetable.fillArray ();
  predicatetable.fillArray ();
  modetable.fillArray ( predicatetable );
  modetable.fillDependencies ( predicatetable, typetable );
}

bool interpret ( AtomString *atom ) {
  if ( parsingbias ) {
    if ( strcmp ( atom->name, "type" ) == 0 )
      return interpret_type ( atom->nrofparams, atom->params );
     // an inttype does not introduce typenames, just tells
     // the maximum number. Displayed are only numbers,
     // facts should use numbers, etc.
    if ( strcmp ( atom->name, "inttype" ) == 0 )
      return interpret_inttype ( atom->nrofparams, atom->params );
    if ( strcmp ( atom->name, "predicate" ) == 0 )
      return interpret_predicate ( atom->nrofparams, atom->params );
    if ( strcmp ( atom->name, "mode" ) == 0 )
      return interpret_mode ( atom->nrofparams, atom->params );
    if ( strcmp ( atom->name, "externmode" ) == 0 )
      return interpret_mode ( atom->nrofparams, atom->params, atom->nrofparams >= 2 ? atom->params[1]->name : (char*) NULL );
    if ( strcmp ( atom->name, "externnamemode" ) == 0 )
      return interpret_mode ( atom->nrofparams - 1, atom->params + 1, atom->params[0]->name );
    if ( strcmp ( atom->name, "key" ) == 0 )
      return interpret_key ( atom->nrofparams, atom->params );
    if ( strcmp ( atom->name, "numberimm" ) == 0 )
      return interpret_numberimm ( atom->nrofparams, atom->params );
    INTERPRET_BOOLOPTION ( atom, showconstantnames );
    INTERPRET_BOOLOPTION ( atom, gainratio );
    INTERPRET_BOOLOPTION ( atom, mdl );
    INTERPRET_BOOLOPTION ( atom, othertest );
    INTERPRET_BOOLOPTION ( atom, shuffle );
    INTERPRET_BOOLOPTION ( atom, constlapunishment );
    INTERPRET_BOOLOPTION ( atom, usefirstattributeforvalidation);
    INTERPRET_BOOLOPTION ( atom, displaydistribution );
    INTERPRET_BOOLOPTION ( atom, prune );
    INTERPRET_BOOLOPTION ( atom, validationtable );

    INTERPRET_INTOPTION ( atom, otherclassleafthreshold );
    INTERPRET_INTOPTION ( atom, leafthreshold );
    INTERPRET_INTOPTION ( atom, crossvalidation );
    INTERPRET_INTOPTION ( atom, maxlookahead );
    INTERPRET_INTOPTION ( atom, lookaheadfraction );
    INTERPRET_DOUBLEOPTION ( atom, lookaheadpunishment );
    INTERPRET_DOUBLEOPTION ( atom, minimpurityincrease );
    INTERPRET_DOUBLEOPTION ( atom, addlap );
    parsingbias = false;
      // no bias definition, going to background knowledge.
      // first put bias in faster access datastructures
    fill ();
    ERR << "[Initializing fact tables]" << endl;
    ERR_F
  }

  return interpret_fact ( atom->name, atom->nrofparams, atom->params );
}

void assertfilledbias () {
  if ( parsingbias )
    fill ();
}



