/***************************************************************************
                          tree.cpp  -  dtfarmer
                             -------------------
    begin                : Sat Mar 25 2000
    copyright            : (C) 2000 by Siegfried Nijssen
    email                : snijssen@cs.leidenuniv.nl
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include "tree_global.h"
#include "type.h"
#include <values.h>
#include <stdio.h>
#include <math.h>
#define SMALLTHRESHOLD 0.000000000000000000000001
#ifdef _DEBUG
#define _DEBUG2
#endif

Node *root;
int treedepth;
Node **nodehistory;
int nodehistorysize;
int *nvar; // number of variables for each type
int *maxnvar;
INTTYPE **varval;
int clasmax; // remembers the default classification

bool determineStop ( KeyFactTables **facts, char *clas, bool force = false, double *impurity = NULL, double *ratio = NULL );

Node::Node ( short int _mode, int size, INTTYPE *_params, void *_additionaldata = NULL ):
  left ( NULL ), right ( NULL ), additionaldata ( _additionaldata ),
  mode ( _mode ), enabled ( true ),
  classn ( NULL ) {
  ModeData *md = modetable.array + mode;
  md->table.table->setCurrentNode ( this );
  md->table.table->init ();
  params = new INTTYPE[size];
  for ( int i = 0; i < size; i++ )
    params[i] = _params[i];
  leftfacts = new KeyFactTables ( modetable.keynr );
  rightfacts = new KeyFactTables ( modetable.keynr );
}

void Node::finish_additionaldata () {
  if ( additionaldata ) {
    ModeData *md = modetable.array + MODE ( this );
    md->table.table->setCurrentNode ( this );
    md->table.table->finish ();
  }
  // the node is ready now, including its tables, so make a copy
  // of the number for displaying purposes if wanted
  if ( displaydistribution ) {
    classn = new unsigned int[modetable.keynr * 2];
    for ( int i = 0; i < modetable.keynr; i++ ) {
      classn[i*2] = ( leftfacts->keyfacttable[i] ? leftfacts->keyfacttable[i]->getSize (): 0 );
      classn[i*2+1] = ( rightfacts->keyfacttable[i] ? rightfacts->keyfacttable[i]->getSize (): 0 );
    }
  }
}

double Node::impurity ( int leftnrkeyfacts, double leftimpurity, double &ratio ) {
  ModeData *md = modetable.array + MODE ( this );
  // this function should not be called for modes which do not
  // have outputs, as it used in refined nodes, which are outputting
  // nodes
  int rightnrkeyfacts = rightfacts->nrkeyfacts (),
      totalnrkeyfacts = rightnrkeyfacts + leftnrkeyfacts;
  if ( rightnrkeyfacts )
    ratio -= rightnrkeyfacts * log ( rightnrkeyfacts );
  double lap = md->table.modefacttable->getLookaheadPunishment ();
  return ( ( leftimpurity * leftnrkeyfacts ) * lap + ( lap - 1 ) * addlap +
           rightfacts->impurity () * rightnrkeyfacts )
         / totalnrkeyfacts;
}

double Node::impurity ( double &ratio ) {
  int rightnrkeyfacts = rightfacts->nrkeyfacts (),
      leftnrkeyfacts = leftfacts->nrkeyfacts (),
      totalnrkeyfacts = rightnrkeyfacts + leftnrkeyfacts;
  ratio -=   ( leftnrkeyfacts ?  ( (double) leftnrkeyfacts ) * log ( leftnrkeyfacts ) : 0 )
           + ( rightnrkeyfacts ? ( (double) rightnrkeyfacts ) * log ( rightnrkeyfacts ) : 0 );
  if ( additionaldata ) {
    ModeData *md = modetable.array + MODE ( this );
    md->table.table->setCurrentNode ( this );
    return ( leftfacts->impurity () * leftnrkeyfacts +
             rightfacts->impurity () * rightnrkeyfacts )
           / totalnrkeyfacts
           + md->table.table->additionalPunishment ();
  }
  return ( leftfacts->impurity () * leftnrkeyfacts +
           rightfacts->impurity () * rightnrkeyfacts )
         / totalnrkeyfacts;
}

char *Node::toString () {
  static char string[255];
  ModeData *md = modetable.array + mode;
  PredicateData *pd = predicatetable.array[md->predicate];
  char *ts = string;

  ts += sprintf ( ts, "%s", pd->name );
  if ( additionaldata ) {
    md->table.table->setCurrentNode ( this );
    ts += sprintf ( ts, ":%s", md->table.table->string () );
  }
  ts += sprintf ( ts, "( " );
  for ( int i = 0; i < md->size; i++ ) {
    switch ( md->params[i].type ) {
      case Plus: ts += sprintf ( ts, "+%i.%i", pd->params[i], params[i] ); break;
      case Min: ts += sprintf ( ts, "-%i.%i", pd->params[i], params[i] ); break;
      case MinEx: ts += sprintf ( ts, "\\%i.%i", pd->params[i], params[i] ); break;
      case Constant:
        if ( showconstantnames )
          ts += sprintf ( ts, "%s", typetable.getConstantName ( pd->params[i], params[i] ) );
        else
          ts += sprintf ( ts, "#%i.%i", pd->params[i], params[i] );
        break;
    }

    if ( i < md->size - 1 )
      ts += sprintf ( ts, ", " );
  }
  ts += sprintf ( ts, " )" );

  if ( displaydistribution ) {
    ts += sprintf ( ts, " <" );
    for ( int i = 0; i < modetable.keynr; i++ )
      ts += sprintf ( ts, " %i", classn[i * 2] );
    ts += sprintf ( ts, " < > " );
    for ( int i = 0; i < modetable.keynr; i++ )
      ts += sprintf ( ts, " %i", classn[i * 2 + 1] );
    ts += sprintf ( ts, " >" );
  }

  return string;
}

Node::~Node () {
  ModeData *md = modetable.array + mode;
  md->table.table->setCurrentNode ( this );
  md->table.table->remove ();
  delete[] params;
  if ( left )
    delete left;
  if ( right )
    delete right;
  if ( leftfacts )
    delete leftfacts;
  if ( rightfacts )
    delete rightfacts;
  if ( classn )
    delete[] classn;
}

NodeList::NodeList ( bool del ):
  firstelt ( NULL ),
  lastelt ( NULL ),
  walkelt ( NULL ) {
  this->del = del;
}

NodeList::~NodeList () {
  NodeListElt *temp;
  while ( firstelt ) {
    temp = firstelt;
    firstelt = firstelt->next;
    if ( del )
      delete temp->node;
    delete temp;
  }
}

void NodeList::addNode ( Node *node ) {
  NodeListElt *newelt = new NodeListElt;
  newelt->next = NULL;
  newelt->prev = lastelt;
  newelt->node = node;
  if ( lastelt ) {
    lastelt->next = newelt;
    lastelt = newelt;
  }
  else
    firstelt = lastelt = newelt;
}

Node *NodeList::remove () {
  NodeListElt *nwe = walkelt->next;

  if ( walkelt->prev )
    walkelt->prev->next = walkelt->next;
  else
    firstelt = walkelt->next;
  if ( walkelt->next )
    walkelt->next->prev = walkelt->prev;
  else
    lastelt = walkelt->prev;

  delete walkelt;

  walkelt = nwe;

  return nwe ? nwe->node : NULL;
}

void NodeList::append ( NodeList *nodelist ) {
  if ( lastelt )
    lastelt->next = nodelist->firstelt;
  else
    firstelt = nodelist->firstelt;
  if ( nodelist->firstelt ) {
    nodelist->firstelt->prev = lastelt;
    lastelt = nodelist->lastelt;
  }

  nodelist->firstelt = nodelist->lastelt = nodelist->walkelt = NULL;
}

void initTree () {
  int i;
  root = NULL;
  treedepth = maxlookahead;
    // treedepth used during creation of the tree to store history,
    // should be larger than necessary in case lookahead needs to
    // be computed.
  nvar = new int[typetable.size];
  maxnvar = new int[typetable.size];

  int *params = akeypredicate->params, size = akeypredicate->size;

  for ( i = 0; i < typetable.size; i++ )
    maxnvar[i] = 0;

  for ( i = 0; i < size; i++, params++ )
    maxnvar[*params]++;

  clasmax = 0;
  int clasmaxval = 0;
  for ( i = 0; i < train_keyfacttables->size; i++ ) {
    if ( train_keyfacttables->keyfacttable[i]->getSize () > clasmaxval ) {
      clasmaxval = train_keyfacttables->keyfacttable[i]->getSize ();
      clasmax = i;
    }
  }
}

void init_history () {
  int *ppd = akeypredicate->params;
  int i, size = akeypredicate->size;
  nodehistory = new Node*[treedepth + 2];

  nodehistorysize = 0;
  for ( i = 0; i < typetable.size; i++ )
    nvar[i] = 0;

  for ( i = 0; i < size; i++, ppd++ )
    nvar[*ppd]++;
}

void remove_history () {
  delete[] nodehistory;
}

void push_to_history ( Node *node ) {
  ModeData *md = modetable.array + MODE ( node );
  PredicateData *pd = predicatetable.array[md->predicate];
  int *ppd = pd->params;
  int nroutput = md->nroutput;
  int *outputvarpos = md->outputvarpos;

  nodehistory[nodehistorysize] = node;
  if ( nodehistorysize > treedepth - maxlookahead )
    treedepth = nodehistorysize + treedepth;
  nodehistorysize++;

  for ( int i = 0; i < nroutput; i++, outputvarpos++ )
    nvar[ppd[*outputvarpos]]++;
}

void dec_mode_counter ( Node *node ) {
  ModeData *md = modetable.array + MODE ( node );
  md->number--;
}

void pop_from_history ( Node *node ) {
  ModeData *md = modetable.array + MODE ( node );
  PredicateData *pd = predicatetable.array[md->predicate];
  int *ppd = pd->params;
  int nroutput = md->nroutput;
  int *outputvarpos = md->outputvarpos;

  nodehistorysize--;

  for ( int i = 0; i < nroutput; i++, outputvarpos++ )
    nvar[ppd[*outputvarpos]]--;
}

void inc_mode_counter ( Node *node ) {
  ModeData *md = modetable.array + MODE ( node );
  md->number++;
}

void init_output_const_params ( INTTYPE *params, ModeData *md ) {
  int tempvarn[typetable.size], i;
  PredicateData *pd = predicatetable.array[md->predicate];
  ModeParamData *mpd = md->params;
  int *ppd = pd->params;

  for ( i = 0; i < typetable.size; i++ )
    tempvarn[i] = 0;

  for ( i = 0; i < md->size; i++, mpd++, ppd++ )
    if ( mpd->type != Plus )
      if ( mpd->type == Constant )
        params[i] = mpd->number;
      else {
        params[i] = nvar[*ppd] + tempvarn[*ppd];
	if ( params[i] >= maxnvar[*ppd] )
	  maxnvar[*ppd] = params[i] + 1;
        tempvarn[*ppd]++;
      }
}

// fill params with the first input params belong with mode
void first_input_params ( INTTYPE *params, ModeData *md ) {
  int inputnr = md->nrinput, *inputvarpos = md->inputvarpos;

  for ( int i = 0; i < inputnr; i++, inputvarpos++ )
    params[*inputvarpos] = 0;
}

// returns true or false! before pos are set to variables which were
// not introduced by the parent node. However, this might not be possible,
// in that case false is returned
bool first_input_params ( int pos, INTTYPE val, INTTYPE *params, INTTYPE *used_var, ModeData *md ) {
  int inputnr = md->nrinput, *inputvarpos = md->inputvarpos;
  ModeParamData *mpd = md->params;

  if ( nvar[mpd[inputvarpos[pos]].number] == used_var[pos] )
    return false;

  for ( int i = 0; i < inputnr; i++, inputvarpos++ )
    if ( i < pos && used_var[i] == 0 )
      return false;
    else
      if ( i == pos )
        params[*inputvarpos] = val;
      else
        params[*inputvarpos] = 0;

  return true;
}

bool next_input_params ( INTTYPE *params, ModeData *md ) {
  PredicateData *pd = predicatetable.array[md->predicate];
  int *inputvarpos = md->inputvarpos, nrinput = md->nrinput, pos = nrinput - 1, *ppd = pd->params;

  while ( params[inputvarpos[pos]] == nvar[ppd[inputvarpos[pos]]] - 1 ) {
    pos--;
    if ( pos < 0 )
      return false;
  }
  params[inputvarpos[pos]]++;

  for ( pos++; pos < nrinput; pos++ )
    params[inputvarpos[pos]] = 0;

  return true;
}

bool next_input_params ( int expos, INTTYPE *params, INTTYPE *used_vars, ModeData *md ) {
  PredicateData *pd = predicatetable.array[md->predicate];
  int *inputvarpos = md->inputvarpos,
      nrinput = md->nrinput,
      pos = nrinput - 1,
      *ppd = pd->params;
  bool b = false;

  // does not stop correctly
  if ( pos == expos ) {
    b = true;
    pos--;
    if ( pos < 0 )
      return false;
  }
  while ( params[inputvarpos[pos]] == ( b ? used_vars[pos] : nvar[ppd[inputvarpos[pos]]] ) - 1 ) {
    pos--;
    if ( pos == expos ) {
      b = true;
      pos--;
    }
    if ( pos < 0 )
      return false;
  }
  params[inputvarpos[pos]]++;

  for ( pos++; pos < nrinput; pos++ )
    if ( pos != expos )
      params[inputvarpos[pos]] = 0;

  return true;
}

void addPossibleExpansions ( NodeList *nodelist, ModeData *md ) {
  INTTYPE params[md->size];

  init_output_const_params ( params, md );

  first_input_params ( params, md );
  do
    nodelist->addNode ( new Node ( md->nr, md->size, params ) );
  while ( next_input_params ( params, md ) );
}

void init_used_vars ( INTTYPE *used_vars, Node *node, ModeData *md ) {
  unsigned int t_used_vars[typetable.size];
  int i, number;
  ModeData *md2 = modetable.array + node->mode;
  ModeParamData *mpd = md->params, *mpd2 = md2->params;
  int *inputvarpos = md->inputvarpos;

  for ( i = 0; i < typetable.size; i++ )
    t_used_vars[i] = 0;

  for ( i = 0; i < md2->size; i++, mpd2++ )
    if ( mpd2->type == Min || mpd2->type == MinEx )
      t_used_vars[mpd2->number]++;

  for ( i = 0; i < md->nrinput; i++ ) {
    number = mpd[inputvarpos[i]].number;
    used_vars[i] = nvar[number] - t_used_vars[number];
  }
}

void addPossibleDependentExpansions ( Node *node, NodeList *nodelist, ModeData *md ) {
  INTTYPE params[md->size];
  INTTYPE used_vars[md->nrinput];
  int i, j, number, *inputvarpos = md->inputvarpos;
  ModeParamData *mpd = md->params;

  init_output_const_params ( params, md );
  init_used_vars ( used_vars, node, md );

  for ( i = 0; i < md->nrinput; i++ ) {
    number = nvar[mpd[inputvarpos[i]].number];
    for ( j = used_vars[i]; j < number; j++ )
      if ( first_input_params ( i, j, params, used_vars, md ) ) 
        do 
          nodelist->addNode ( new Node ( md->nr, md->size, params ) );
        while ( next_input_params ( i, params, used_vars, md ) );
  }
}

bool test_for_inputs ( ModeData *md ) {
  PredicateData *pd = predicatetable.array[md->predicate];
  int i, size = md->nrinput, *ppd = pd->params, *inputvarpos = md->inputvarpos;

  for ( i = 0; i < size; i++, inputvarpos++ )
    if ( nvar[ppd[*inputvarpos]] == 0 )
      return false;

  return true;
}

NodeList *determinePossibleExpansions () {
  NodeList *nodelist = new NodeList;
  ModeData *md = modetable.array;

  for ( int i = 0; i < modetable.size; i++, md++ )
    if ( !md->key && md->number && test_for_inputs ( md ) )
      addPossibleExpansions ( nodelist, md );

  return nodelist;
}

NodeList *determinePossibleDependentExpansions ( Node *node ) {
  NodeList *nodelist = new NodeList;
  ModeData *md = modetable.array + node->mode, *md2;
  int depmodsize = md->depmodsize, *depmod = md->dependentmodes;;

  for ( int i = 0; i < depmodsize; i++, depmod++ ) {
    md2 = modetable.array + *depmod;
    if ( md2->number && test_for_inputs ( md2 ) )
      addPossibleDependentExpansions ( node, nodelist, md2 );
  }

  return nodelist;
}

void copy_key_var_data ( INTTYPE *value ) {
  int *params = akeypredicate->params, size = akeypredicate->size;
  int varn[typetable.size], i;

  for ( i = 0; i < typetable.size; i++ )
    varn[i] = 0;

  for ( i = 0; i < size; i++, params++ )
    varval[*params][varn[*params]++] = value[i];
}

void evaluate_node ( Node *node, bool &result, ModeFactCoors *&resultcoors ) {
  ModeData *md = modetable.array + MODE ( node );
  int i, j, size = md->size;
  INTTYPE coors[md->size - md->nroutput];
  ModeParamData *mpd = md->params;
  INTTYPE *npd = node->params;

  for ( i = 0, j = 0; i < size; i++, mpd++, npd++ )
    if ( mpd->type == Constant )
      coors[j++] = mpd->number;
    else
      if ( mpd->type == Plus )
        coors[j++] = varval[mpd->number][*npd];

  md->table.table->setCurrentNode ( node );
  if ( md->outputs ) {
    resultcoors = md->table.modefacttable->getTrueFacts ( coors );
    result = resultcoors->empty ();
  }
  else {
    resultcoors = NULL;
    result = md->table.facttable->getFact ( coors );
  }
}

// value contains output values for the given node
void copy_node_var_data ( Node *node, INTTYPE *value ) {
  ModeData *md = modetable.array + MODE ( node );
  PredicateData *pd = predicatetable.array[md->predicate];
  int outputnr = md->nroutput, *outputvarpos = md->outputvarpos, i;
  int *ppd = pd->params;
  INTTYPE *npd = node->params;

  for ( i = 0; i < outputnr; i++, outputvarpos++ )
    varval[ppd[*outputvarpos]][npd[*outputvarpos]] = value[i];
}

// return false: not continue search
bool evaluate_leaf ( int crn, INTTYPE *value, NodeList *nodelist, NodeList *provednodelist ) {
  Node *cn = nodelist->first ();
  bool result, res = true;
  ModeFactCoors *modefactcoors;
  BaseTable *bt;

  while ( cn ) {
    evaluate_node ( cn, result, modefactcoors );
    if ( modefactcoors )
      delete modefactcoors; // coordinates not necessary, remove view
    if ( result ) {
      if ( !cn->leftfacts->keyfacttable[crn] )
        cn->leftfacts->keyfacttable[crn] = new KeyFactTable ( akeypredicate->size, false );
      bt = modetable.array[MODE ( cn )].table.table;
      cn->leftfacts->keyfacttable[crn]->addFact ( value, bt->determineAdditionalData () );
      provednodelist->addNode ( cn );
      cn = nodelist->remove ();
    }
    else {
      res = false;
      cn = nodelist->next ();
    }
  }

  return !res;
}

// return false: not continue search
bool evaluate_node ( int depth, int crn, INTTYPE *value, NodeList *nodelist, NodeList *provednodelist ) {
  if ( nodehistorysize == depth )
    return evaluate_leaf ( crn, value, nodelist, provednodelist );
  Node *node = nodehistory[depth];
  ModeFactCoors *modefactcoors;
  bool result;
  INTTYPE *modevalue;

  evaluate_node ( node, result, modefactcoors );

  if ( modefactcoors ) {
    while ( modevalue = modefactcoors->value () ) {
      // put the modefactors on the varval history
      copy_node_var_data ( node, modevalue );
      if ( !evaluate_node ( depth + 1, crn, value, nodelist, provednodelist ) ) {
	delete modefactcoors;
	return false;
      }
    }
    delete modefactcoors;
  }
  else
    if ( result )
      return evaluate_node ( depth + 1, crn, value, nodelist, provednodelist );
  
  return true;
}

void add_value_to_unproved ( int cnr, INTTYPE *value, NodeList *unprovednodelist ) {
  Node *node = unprovednodelist->first ();
  KeyFactTable **kft;
  BaseTable *bt;

  while ( node ) {
    bt = modetable.array[MODE ( node )].table.table;
    bt->setCurrentNode ( node );
    kft = node->rightfacts->keyfacttable + cnr;
    if ( !*kft )
      *kft = new KeyFactTable ( akeypredicate->size, false );
    (*kft)->addFact ( value, bt->determineAdditionalData () );
    node = unprovednodelist->next ();
  }
}

void evaluate_fact ( int cnr, void *additionaldata_class, INTTYPE *value, void *additionaldata_value, void *extraparams ) {
  NodeList *nodes = (NodeList*) extraparams, *provednodes = new NodeList;

  copy_key_var_data ( value );

  evaluate_node ( 0, cnr, value, nodes, provednodes );

  // nodes contains only those nodes which have not been proved
  add_value_to_unproved ( cnr, value, nodes );

  nodes->append ( provednodes );
  // concatenate provednodes to nodes, for next value all should
  // be present. provednodes have been removed in order to avoid
  // them from being reconsidered for other non-key output variables
  // it already has been shown that they are true

  delete provednodes;
}

void evaluate_facts ( KeyFactTables *facts, NodeList *nodes ) {
  facts->iterate ( evaluate_fact, (void*) nodes );
}

void init_var_data () {
  varval = new INTTYPE *[typetable.size];
  for ( int i = 0; i < typetable.size; i++ )
    if ( maxnvar[i] )
      varval[i] = new INTTYPE[maxnvar[i]];
    else
      varval[i] = NULL;
}

void remove_var_data () {
  for ( int i = 0; i < typetable.size; i++ )
    if ( varval[i] )
      delete[] varval[i];
  delete[] varval;
}

double determineExpansion ( Node **nodeptr, KeyFactTables **facts, char *clas, NodeList *newnodes, int refinelevel, double *ratio );

double refineNode ( Node *node, int refinelevel, double *ratio ) {
  double impurity;

  if ( determineStop ( &(node->leftfacts), &(node->leftclass ), false, &impurity, ratio ) ) {
    dec_mode_counter ( node );
    push_to_history ( node );

    impurity = determineExpansion ( &(node->left), &(node->leftfacts), &(node->leftclass), determinePossibleDependentExpansions ( node ), refinelevel + 1, ratio );

    pop_from_history ( node );
    inc_mode_counter ( node );
  }

  return impurity;
}

void printDistr ( KeyFactTables *kfts ) {
  for ( int i = 0; i < kfts->size; i++ )
    if ( kfts->keyfacttable[i] )
      cout << kfts->keyfacttable[i]->getSize () << " ";
    else
      cout << "0 ";
}

// key facttables: the set of facts which must be split by the
// test; it is used to determine whether a node needs further
// refinement because it has outputs and does not split enough
// refinelevel: if refining, the depth level we have currently refined
Node *determineBestNode ( KeyFactTables *kfts, NodeList *nodelist, int refinelevel = 0, double *chosenimpurity = NULL, double *chosenratio = NULL ) {
  Node *node = nodelist->first ();
  ModeData *md;
  double minimpurity, impurity, leftimpurity,
         ratiovalue, minratiovalue, minratio, ratio,
         origimpurity = kfts->impurity ();
  bool first = true;
  Node *minnode;
  int s = kfts->nrkeyfacts ();

#ifdef _DEBUG
  static int cnrs = 0;
  int cnr = cnrs;
  cnrs++;

  cout << cnr << " Impurity (current): " << origimpurity << endl;
#endif

  while ( node ) {
    node->finish_additionaldata ();
      // the node should take its final shape NOW!
    md = modetable.array + node->mode;

    if ( md->outputs && node->isRightLookaheadFraction ( kfts ) && refinelevel < maxlookahead ) {
      int leftnrkeyfacts = node->leftfacts->nrkeyfacts ();
      ratiovalue = 0;
      leftimpurity = refineNode ( node, refinelevel, &ratiovalue );
        // try to refine
      impurity = node->impurity ( leftnrkeyfacts, leftimpurity, ratiovalue );
        // use impurity of refinement, punishing for larger depth

#ifdef _DEBUG
      cout << cnr << " " << node->toString ()
           << " " << impurity << " (refined)"
           << " left: " << leftimpurity
           << " right: " << node->rightfacts->impurity ();
#endif
    }
    else {
      ratiovalue = 0;
      impurity = node->impurity ( ratiovalue );

#ifdef _DEBUG
      cout << cnr << " " << node->toString ()
           << " " << impurity
           << " left: " << node->leftfacts->impurity () << " (";
      printDistr ( node->leftfacts );
      cout << ") right: " << node->rightfacts->impurity () << " (";
      printDistr ( node->rightfacts );
      cout << ")";
#endif
    }

    if ( gainratio ) {
      if ( origimpurity - impurity < SMALLTHRESHOLD )
        ratio = 0;
      else {
        double t = ratiovalue / ( (double) s ) + log ( s );
#ifdef _DEBUG
	cout << ratiovalue << " r " << t << " t ";
#endif
        if ( t < SMALLTHRESHOLD )
          ratio = 0;
        else
          ratio = ( origimpurity - impurity ) / t;
      }
      if ( ratio > MAXDOUBLE ) {
        cout << "ORIG " << origimpurity << " IMP " << impurity << " RATVAL " << ratiovalue << " S " << s << " SLOGS " << ((double)s)*log(s) << endl;
      }
    }
    else
      ratio = origimpurity - impurity;

#ifdef _DEBUG
    cout << ratio << endl;
#endif

    if ( first ) {
      minnode = node;
      minimpurity = impurity;
      minratio = ratio;
      minratiovalue = ratiovalue;
      first = false;
    }
    else {
      if ( ratio > minratio ) {
        delete minnode; // delete the previous minimum
        minnode = node;
        minratio = ratio;
        minimpurity = impurity;
        minratiovalue = ratiovalue;
      }
      else
        delete node; // delete this node, can not be the minimum
    }
    node = nodelist->remove ();
  }
  
  delete nodelist;

#ifdef _DEBUG
  cout << cnr << " Current best node: " << minnode->toString () << " (" << minratio << "," << minimpurity << ")" << endl;
#endif
  if ( minratio < minimpurityincrease ) {
     // no enhancement
    delete minnode;
    minnode = NULL;
    *chosenimpurity = origimpurity;
    *chosenratio = - ( (double) s ) * log ( s );
  }
  else {
    *chosenimpurity = minimpurity;
    *chosenratio = minratiovalue;
  }

#ifdef _DEBUG
  if ( minnode ) {
    cout << cnr << " Chosen node: " << minnode->toString () << " (" << minratio << "," << *chosenimpurity << ")" << endl;
  }
  else
    cout << cnr << " No increase for best node, no node chosen" << endl;
#endif

  return minnode;
}

// stopping criterion
bool determineStop ( KeyFactTables **facts, char *clas, bool force = false , double *impurity = NULL, double *ratio = NULL ) {
  int size, maxs = -1, sum = 0;
  char maxi;
  KeyFactTable *kft;

  for ( int i = 0; i < (*facts)->size; i++ ) {
    kft = (*facts)->keyfacttable[i];
    if ( kft )
      size = kft->getSize ();
    else
      size = 0;
    sum += size;
    if ( size > maxs ) {
      maxs = size;
      maxi = i;
    }
  }

  if ( sum < leafthreshold ||
       maxs >= sum - otherclassleafthreshold /* (almost) all the same class */ ||
       force ) {
    if ( sum == 0 )
      *clas = clasmax;
    else
      *clas = maxi;
    if ( impurity ) {
      *impurity = (*facts)->impurity ();
      *ratio = - (*facts)->nrkeyfacts () * log ( (*facts)->nrkeyfacts () );
    }
    if ( *facts != train_keyfacttables ) {
      delete *facts;
      *facts = NULL;
    }
    return false;
  }
  return true;
}

// compute the best test for a set of facts; the resulting
// test will be stored in *nodeptr, if testing not reasonable or
// possible (run out of modes, eg): store default class in *clas
// *facts will be emptied in case of no next test.
// provide the set of candidate tests
// returns the impurity after the selected test, or the impurity of
// the facts set if no test was chosen
double determineExpansion ( Node **nodeptr, KeyFactTables **facts, char *clas, NodeList *newnodes, int refinelevel = 0, double *ratio = NULL ) {
  bool b;
  double impurity;
  Node *bestnode;
  double dummyratio = 0;
  if ( !ratio )
    ratio = &dummyratio;

  if ( !newnodes->first () ) { // no possible new test, just stop
    determineStop ( facts, clas, true, &impurity, ratio );
    return impurity;
  }

  init_var_data ();

  evaluate_facts ( *facts, newnodes );

  remove_var_data ();
  bestnode = determineBestNode ( *facts, newnodes, refinelevel, &impurity, ratio );

  if ( !bestnode ) {
    // no test can be found that splits up the examples satisfactionary
    // (eg: the examples are contradictory)
    determineStop ( facts, clas, true, &impurity, ratio );
    return impurity;
  }

  *nodeptr = bestnode;

  b = determineStop ( &( (*nodeptr)->rightfacts ), &( (*nodeptr)->rightclass ) );

  if ( (*nodeptr)->leftfacts )
    b = determineStop ( &( (*nodeptr)->leftfacts ), &( (*nodeptr)->leftclass ) ) || b;
  else
    if ( (*nodeptr)->left )
      b = (*nodeptr)->left->enabled || b;
  
  if ( !b )
    (*nodeptr)->enabled = false;

  return impurity;
}

void expandTree ( Node *node ) {
  if ( node->enabled ) {
    bool b;

    dec_mode_counter ( node );
    push_to_history ( node );
    if ( node->left ) {
      if ( node->leftfacts ) {
        delete node->leftfacts;
        node->leftfacts = NULL;
      }
      expandTree ( node->left );
      b = node->left->enabled;
    }
    else
      if ( node->leftfacts ) {
        determineExpansion ( &(node->left), &(node->leftfacts), &(node->leftclass ), determinePossibleExpansions () );
        if ( node->left )
          b = node->left->enabled;
        delete node->leftfacts;
        node->leftfacts = NULL;
      }
      else
        b = false;
    pop_from_history ( node );


    if ( node->right ) {
      if ( node->rightfacts ) {
        delete node->rightfacts;
        node->rightfacts = NULL;
      }
      expandTree ( node->right );
      b = node->right->enabled || b;
    }
    else
      if ( node->rightfacts ) {
        determineExpansion ( &(node->right), &(node->rightfacts), &(node->rightclass ), determinePossibleExpansions () );
        // remember: node->right may be changed
        if ( node->right )
          b = node->right->enabled || b;
        delete node->rightfacts;
        node->rightfacts = NULL;
      }

    inc_mode_counter ( node );

    node->enabled = b;
  }
}

bool expandTree () {
  init_history ();
  char clas; // clas parameter of determinExpansion is filled
             // with the class if no expansion possible. When
             // starting, this should not be possible

  if ( root )
    expandTree ( root );
  else {
    determineExpansion ( &root, &train_keyfacttables, &clas, determinePossibleExpansions () );
  }
  treedepth++;

  remove_history ();

#ifdef _DEBUG2
  printTree ();
#endif

  if ( root )
    return root->enabled;
  else
    // no tree was built, as not enough information to reasonably split anywhere
    return false;
}

// returns classification, -1 if no equals
int safePruneTree ( Node *node ) {
  int class1, class2;

  if ( node->left ) {
    class1 = safePruneTree ( node->left );
    if ( class1 != -1 ) {
      delete node->left;
      if ( node->leftfacts )
        delete node->leftfacts;
      node->leftfacts = NULL;
      node->leftclass = class1;
      node->left = NULL;
    }
  }
  else
    class1 = node->leftclass;
  if ( node->right ) {
    class2 = safePruneTree ( node->right );
    if ( class2 != -1 ) {
      delete node->right;
      if ( node->rightfacts )
        delete node->rightfacts;
      node->rightfacts = NULL;
      node->rightclass = class2;
      node->right = NULL;
    }
  }
  else
    class2 = node->rightclass;
  if ( class1 == class2 )
    return class1;
  else
    return -1;
}

void createTree () {
  int c = 1;
  initTree ();
  do {
    ERR << "Iteration " << c++;
    ERR_F
  }
  while ( expandTree () );
  if ( root && prune )
    safePruneTree ( root );
}

void printSpaces ( int n ) {
  for ( int i = 0; i < n; i++ )
    cout << " ";
}

void printTree ( Node *node, int depth ) {
  cout << node->toString () << "?" << endl;
  printSpaces ( depth + 2 );
  cout << "Yes: ";
  if ( node->left )
    printTree ( node->left, depth + 2 );
  else
    if ( node->leftfacts )
      cout << "Not finished (" << node->leftfacts->nrkeyfacts () << " facts)" << endl;
    else
      cout << predicatetable.array[keynrtopnr[node->leftclass]]->name << endl;
  printSpaces ( depth + 2 );
  cout << "No:  ";
  if ( node->right )
    printTree ( node->right, depth + 2 );
  else
    if ( node->rightfacts )
      cout << "Not finished (" << node->rightfacts->nrkeyfacts () << " facts)" << endl;
    else
      cout << predicatetable.array[keynrtopnr[node->rightclass]]->name << endl;
}

void printTree () {
  if ( root )
    printTree ( root, 0 );
}

void removeTree () {
  if ( root ) {
    delete root;
    delete[] nvar;
    delete[] maxnvar;
  }
}
